/* Copyright (c) 2013 APDM
*
* mobility_clinic_main.js
*/

// Set the paths for server host and port
/* global utils:false, longitudinalPlot:false */
var serverInfo = utils.parseURLParams(window.location.href);

// Values for the "Back" and "Next" buttons
var back = null;
var next = null;

// Global name for unassigned subjects
var UNASSIGNED_SUBJECT = "Unnamed";

// Start a poll on the server to get status information
var polling = false;
var serverStatus = null;
// Start the polling
setInterval(pollServer, 2000);

// Enum for error types
var errorType = {
    DOCK_ERROR: 0,
    MONITOR_ERROR: 1,
    CONFIGURATION_ERROR: 2,
    FILE_ERROR: 3
};

// Is the error page being displayed?
var errorPageOpen = false;

// Constants to use when posting state back to server
// IMPORTANT! These must match the names in
// com.apdm.mobilitylab.views.ClinicView.ClinicUIState
var READY_TO_START = "READY_TO_START";
var WAITING_FOR_ATTACH = "WAITING_FOR_ATTACH";
var WAITING_FOR_DOCK = "WAITING_FOR_DOCK";
var RUNNING_TEST = "RUNNING_TEST";
var WAITING_FOR_REDOCK = "WAITING_FOR_REDOCK";
var WAITING_FOR_ANALYSIS = "WAITING_FOR_ANALYSIS";
var READING_RESULTS = "READING_RESULTS";
var DOCK_ERROR_PAGE = "DOCK_ERROR_PAGE";
var CONFIGURATION_ERROR_PAGE = "CONFIGURATION_ERROR_PAGE";
var FILE_ERROR_PAGE = "FILE_ERROR_PAGE";

// List of metrics that are to be included in the reports
var gaitTrackMetrics = {
    gaitSpeed: {name: "Gait Speed", type: "mean"},
    strideLength: {name: "Stride Length", type: "mean"},
    strideLengthStd: {name: "Stride Length", type: "std"},
    cadence: {name: "Cadence", type: "mean"},
    stepDuration: {name: "Step Duration", type: "mean"},
    heelstrikeLeft: {name: "Pitch at Heel Strike L", type: "mean"},
    heelstrikeRight: {name: "Pitch at Heel Strike R", type: "mean"},
    toeoffLeft: {name: "Pitch at Toe Off L", type: "mean"},
    toeoffRight: {name: "Pitch at Toe Off R", type: "mean"},
};

// Poll the server to check for changes in the server state.
function pollServer() {
    if (polling === false) {
        polling = true;
        $.post("http://" + serverInfo.serverHost + ":" +
               serverInfo.serverPort + "/mobility_clinic/",
               {
                   requestType: "mobilityclinic",
                   type: "status",
               },
               function(d) {
                   serverStatus = d;
                   console.log(serverStatus);
                   polling = false;
                   if (serverStatus.clinicViewState === "TEST_DOCK_ERROR" &&
                       errorPageOpen === false) {
                       loadErrorPage(errorType.DOCK_ERROR);
                   } else if (serverStatus.clinicViewState === "TEST_CONFIGURATION_ERROR" &&
                       errorPageOpen === false) {
                           loadErrorPage(errorType.CONFIGURATION_ERROR);
                   } else if (serverStatus.clinicViewState === "TEST_FILE_ERROR" &&
                              errorPageOpen === false) {
                       loadErrorPage(errorType.FILE_ERROR);
                   }
               },
               "json");
    }
}

// Display an error page, and then wait until the server is ready to
// start a new test.
function loadErrorPage(error) {
    errorPageOpen = true;
    if (error === errorType.CONFIGURATION_ERROR) {
        postUIState(CONFIGURATION_ERROR_PAGE);
        $('#switchable-div').load('config_error.html', pageLoaded);
    } else if (error === errorType.DOCK_ERROR) {
        postUIState(DOCK_ERROR_PAGE);
        $('#switchable-div').load('error.html', pageLoaded);
    } else if (error === errorType.FILE_ERROR) {
        postUIState(FILE_ERROR_PAGE);
        $('#switchable-div').load('file_error.html', pageLoaded);
    } else {
        errorPageOpen = false;
        return;
    }


    function pageLoaded() {
        // Handle the "Next" button if this is a file error
        if (error === errorType.FILE_ERROR) {
            d3.select(".next_button").attr("disabled", "disabled")
                .on("click", loadStartPage);
        }
        // Go back to start page when the error clears
        var timer = setInterval(statusCheck, 500);
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                serverStatus.clinicViewState === "TEST_WAITING_TO_RUN") {
                errorPageOpen = false;
                if (error === errorType.FILE_ERROR) {
                    clearInterval(timer);
                    d3.select(".next_button").attr("disabled", null);
                    return;
                }
                clearInterval(timer);
                loadStartPage();
            }
        }
    }
}


// Load the initial page.
loadStartPage();

function loadStartPage() {
    postUIState(WAITING_FOR_DOCK);
    $('#switchable-div')
        .load('launch.html',
              function() {
                  d3.select("#start-button")
                      .on("click", function() {
                          if (serverStatus !== null &&
                              (serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                               serverStatus.clinicViewState === "TEST_WAITING_TO_RUN")) {
                              attachSensorsPage();
                          } else if (serverStatus === null ||
                                     serverStatus.clinicViewState === "TEST_WAITING_FOR_DOCK") {
                              redockPage();
                          }
                      });
              });
}

function redockPage() {
    $('#switchable-div')
        .load('dock.html', pageLoaded);

    function pageLoaded() {
        back = loadStartPage;
        next = attachSensorsPage;
        navButtonHandlers();

        var timer = setInterval(statusCheck, 500);
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_READY_TO_START" ||
                serverStatus.clinicViewState === "TEST_WAITING_TO_RUN") {
                clearInterval(timer);
                attachSensorsPage();
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

function attachSensorsPage() {
    $('#switchable-div')
        .load('attach.html',
              pageLoaded);

    function pageLoaded() {
        var timer = setInterval(statusCheck, 500);
        back = loadStartPage;
        next = function() {
            clearInterval(timer);
            walkPage();
        };
        navButtonHandlers();

        postUIState(WAITING_FOR_ATTACH);
        d3.select(".next_button").attr("disabled", "disabled");
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_RUNNING") {
                d3.select(".next_button").attr("disabled", null);
            } else if (serverStatus.clinicViewState === "TEST_WAITING_FOR_DOCK") {
                clearInterval(timer);
                redockPage();
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

var seconds = 120;
var countdownTimer = null;
function walkPage() {
    console.log("walkPage");
    $('#switchable-div').load('walk.html', pageLoaded);

    function pageLoaded() {
        back = attachSensorsPage;
        next = dockSensorsPage;
        navButtonHandlers();
        postUIState(RUNNING_TEST);
        // Handle the timer button
        var timerButton = d3.select("#time-button");
        var timerText = d3.select(".countdown_timer");
        var seconds = 120;
        var timerStarted = false;

        timerButton.on("click", function() {
            if (!timerStarted) {
                countdownTimer = setInterval(updateTimer, 1000);
                d3.select(this).text("Reset");
                timerStarted = true;
            } else if (errorPageOpen === true) {
                clearInterval(countdownTimer);
                countdownTimer = null;
            } else {
                clearInterval(countdownTimer);
                countdownTimer = null;
                d3.select(this).text("Start");
                timerText.style("color", "black").text("120");
                seconds = 120;
                timerStarted = false;
            }
        });

        function updateTimer() {
            seconds--;
            timerText.text(seconds);
            if (seconds < 0)
                timerText.style("color", "red");
        }
    }
}

function dockSensorsPage() {
    console.log("dockSensorsPage");
    $('#switchable-div').load('dock.html', pageLoaded);

    // Kill the countdown timer if it's still running
    if (countdownTimer !== null) {
        clearInterval(countdownTimer);
        seconds = 120;
    }

    postUIState(WAITING_FOR_DOCK);
    function pageLoaded() {
        // Move on if the sensors are docked
        if (serverStatus.clinicViewState === "TEST_PROCESSING" ||
            serverStatus.clinicViewState === "TEST_COMPLETE" ||
            serverStatus.clinicViewState === "TEST_COPYING_FILES" ||
            serverStatus.clinicViewState === "TEST_CONVERTING_FILES") {
            analyzeResultsPage();
            return;
        }

        var timer = setInterval(statusCheck, 500);
        
        function statusCheck() {
            if (serverStatus.clinicViewState === "TEST_PROCESSING" ||
            serverStatus.clinicViewState === "TEST_COMPLETE" ||
            serverStatus.clinicViewState === "TEST_COPYING_FILES" ||
            serverStatus.clinicViewState === "TEST_CONVERTING_FILES") {
                clearInterval(timer);
                analyzeResultsPage();
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

function analyzeResultsPage() {
    console.log("analyzeResultsPage");
    $('#switchable-div').load('analyze.html', pageLoaded);
    
    postUIState(WAITING_FOR_ANALYSIS);

    function pageLoaded() {
        var timer = setInterval(statusCheck, 500);

        function statusCheck() {
            var infoText = d3.select(".instruction_big").select("p");
            if (serverStatus.clinicViewState === "TEST_COPYING_FILES") {
                infoText.text("Copying data from monitors.");
            } else if (serverStatus.clinicViewState === "TEST_CONVERTING_FILES") {
                infoText.text("Converting monitor data.");
            } else if (serverStatus.clinicViewState === "TEST_PROCESSING") {
                infoText.text("Analyzing monitor data.");
            } else if (serverStatus.clinicViewState === "TEST_PROCESSED") {
                clearInterval(timer);
                $.post("http://" + serverInfo.serverHost + ":" +
                       serverInfo.serverPort + "/mobility_clinic/",
                       {
                           requestType: "mobilityclinic",
                           type: "get_study",
                       },
                       function(d) {
                           reportPage(d);
                       },
                       "json");
            } else if (errorPageOpen === true) {
                clearInterval(timer);
            }
        }
    }
}

function reportPage(studyData, publicId) {
    console.log("reportPage");
    $('#switchable-div').load('report.html', pageLoaded);

    postUIState(READING_RESULTS);

    function pageLoaded() {
        next = loadStartPage;
        navButtonHandlers();

        d3.select(".add_subject_button")
            .on("click", function() { return addSubjectsPage(studyData); });
        d3.select(".view_history_button").attr("disabled", null)
            .on("click", function() {
                return showHistoryPage(studyData, publicId);
            });
        if (!publicId || publicId === UNASSIGNED_SUBJECT) {
            d3.select(".view_history_button")
                .attr("disabled", "disabled");
        }

        buildSubjectSelect(studyData, publicId);
        buildResultsTable(findCurrentTrial(studyData, publicId));
    }
}

// Find the current trial for a given subject.
function findCurrentTrial(studyData, publicId) {
    if (!publicId) {
        publicId = UNASSIGNED_SUBJECT;
    }
    var trials = findAllTrialsForSubject(studyData, publicId);
    trials.sort(function (a, b) { return d3.descending(a.date, b.date); });
    return trials[0];
}

// Get all the trials for a subject by publicId
function findAllTrialsForSubject(studyData, publicId) {
    var subject = findSubjectInStudy(studyData, publicId);
    if (subject !== null) {
        return subject.children[0].children;
    }
    return null;
}

// Find the subject in study data given the publicId
function findSubjectInStudy(studyData, publicId) {
    for (var i = 0; i < studyData.children.length; i++) {
        if (studyData.children[i].publicId === publicId)
            return studyData.children[i];
    }
    return null;
}

// Build the drop down and buttons for the subject selection
function buildSubjectSelect(studyData, publicId) {
    var select = d3.select("select")
        .on("change", function() { return subjectSelected(this, studyData); });
    select.selectAll("option")
        .data(studyData.children)
        .enter()
        .append("option")
        .each( function(d) {
            var opt = d3.select(this);
            opt.attr("value", d.publicId + ":" + d.id)
                .text(d.publicId);
            if (d.publicId === publicId)
                opt.attr("selected", "selected");
        });


}
 
// Add the entries that we can to the results table.
function buildResultsTable(currTrial) {
    var resultsMap = d3.nest()
        .key(function(d) { return d.metricName; })
        .map(currTrial.metrics, d3.map);

    for (var prop in gaitTrackMetrics) {
        if (resultsMap[gaitTrackMetrics[prop].name]) {
            d3.select("#" + prop)
                .text(resultsMap[gaitTrackMetrics[prop].name][0][gaitTrackMetrics[prop].type].toPrecision(3));
        }
    }
}

function subjectSelected(option, studyData) {
    console.log("subject selected: " + this.value);
    var values = option.value.split(":");
    $.post("http://" + serverInfo.serverHost + ":" +
           serverInfo.serverPort + "/mobility_clinic/",
           {
               requestType: "mobilityclinic",
               type: "change_subject",
               publicId: values[0],
               id: values[1],
           });

    // Handle "View History" button
    var btn = d3.select(".view_history_button");
    if (values[0] !== UNASSIGNED_SUBJECT) {
        btn.attr("disabled", null)
            .on("click", function() {
                return showHistoryPage(studyData, values[0]);
            });
    } else {
        btn.attr("disabled", "disabled");
    }
}

function addSubjectsPage(studyData) {
    console.log("addSubjectsPage");
    $('#switchable-div').load('add_subject.html', pageLoaded);

    function pageLoaded() {
        back = function() { return reportPage(studyData); };
        navButtonHandlers();

        d3.select(".save_button").on("click", saveSubject);
    }

    function saveSubject() {
        var publicId = d3.select("#public-id").node().value;
        var yearOfBirth = d3.select("#year-of-birth").node().value;
        var height = d3.select("#height").node().value;
        var control = d3.select("#control").node().checked;

        $.post("http://" + serverInfo.serverHost + ":" +
               serverInfo.serverPort + "/mobility_clinic/",
               {
                   requestType: "mobilityclinic",
                   type: "new_subject",
                   publicId: publicId,
                   yearOfBirth: yearOfBirth,
                   height: height,
                   control: control,
               },
               function(d) {
                   reportPage(d, publicId);
               },
               "json");
    }
}

function postUIState(state) {
    console.log("Posting state: " + state);
    $.post("http://" + serverInfo.serverHost + ":" +
           serverInfo.serverPort + "/mobility_clinic/",
           {
               requestType: "mobilityclinic",
               type: "ui_state",
               uiState: state,
           });
}

function navButtonHandlers() {
    d3.select(".back_button").on("click", back);
    d3.select(".next_button").on("click", next);
}


//
// Reporting
//

function showHistoryPage(studyData, publicId) {
    console.log("showHistoryPage");
    $('#switchable-div').load('view_history.html', pageLoaded);

    function pageLoaded() {
        next = loadStartPage;
        back = function() { return reportPage(studyData, publicId); }
        navButtonHandlers();

        // Get the subject trials, then send them to plot.
        var trials = findAllTrialsForSubject(studyData, publicId);
        var plotData = formatLongitudinalPlotData(trials, gaitTrackMetrics);

        // Build plots for each metric.
        var div = d3.select(".history_plots");
        plotData.forEach(function(d) {
            var plot = longitudinalPlot().width(800).height(200)
                .x(function(d) { return d.date; })
                .y(function(d) { return d.value; })
                .type(function(d) { return d.condition; });

            var plotDiv = div.append("div").attr("id", "longitudinal-plot");
            plot.xAxisLabel("Time/Date").title(d.name);
            plotDiv.datum(d.values).call(plot);
        });
    }
}

// Build an array where each entry in the array is an object
// containing all the data across trials for a given metric.
function formatLongitudinalPlotData(trials, metricSet) {
    var plotData = [];
    trials = getDateAndMetricsFromTrials(trials, metricSet);
    trials.forEach(function(d) {
        d.metrics = d3.nest()
            .key(function(e) { return e.metricName; })
            .map(d.metrics, d3.map);
    });

    for (var prop in metricSet) {
        var metricData = { name: metricSet[prop], values: [] };
        trials.forEach(function(d) {
            if (d.metrics[metricSet[prop]]) {
                metricData.values
                    .push({ value: d.metrics[metricSet[prop]][0].mean,
                            date: utils.getDateFromString(d.date) });
            }
        });
        plotData.push(metricData);
    }
    return plotData;
}

// Given an array of trials, return an array of objects containing the
// trial date, condition, and an array of just the metrics we care
// about.
function getDateAndMetricsFromTrials(trials, metricSet) {
    var trialData = trials.map(function(d) {
        return { date: d.date,
                 metrics: getMetricSetMetrics(d.metrics, metricSet) };
    });
    return trialData;
}

// Pull just the metrics we care about out of an array of trial metrics.
function getMetricSetMetrics(metrics, metricSet) {
    var results = [];
    metrics.forEach(function(d) {
        for (var prop in metricSet) {
            if (metricSet[prop] === d.metricName) {
                results.push(d);
            }
        }
    });
    return results;
}
